! file src/solver_module.f90
!
! Copyright 2009,2010 Dalton Harvie (daltonh@unimelb.edu.au)
! 
! This file is part of arb finite volume solver, referred to as `arb'.
! 
! arb is a software package designed to solve arbitrary partial
! differential equations on unstructured meshes using the finite volume
! method.  Primarily it consists of fortran source code, perl source
! code and shell scripts.  arb replies on certain third party software
! to run, most notably the computer algebra system maxima
! <http://maxima.sourceforge.net/> which is released under the GNU GPL.
! 
! The copyright of arb is held by Dalton Harvie.
! 
! arb is released under the GNU GPL.  arb is free software: you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License (version 3) as published by the Free Software Foundation.
! You should have received a copy of the GNU General Public License
! along with arb (see file license/gpl.txt after unpacking).  If not,
! see <http://www.gnu.org/licenses/>.
! 
! arb is distributed in the hope that it will be useful, but WITHOUT
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
! FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
! for more details.
! 
! For full details of arb's license see the license directory.
! 
! The current homepage for the arb finite volume solver project is
! <http://www.chemeng.unimelb.edu.au/people/staff/daltonh/downloads/arb>.
!
!-------------------------------------------------------------------------
module solver_module

!-----------------------------------------------------------------
contains

!-----------------------------------------------------------------

subroutine newtsolver(error,newtres)

! solves equations using newton's method

use general_module
use equations_module
implicit none
integer :: m, error, p, n, ns
double precision :: alf, lambda, lambdamin, newtres, newtresold
double precision, dimension(:), allocatable, save :: phiold, delphi ! single dimension independent-sized variables for newton proceedure
logical :: backstepping
logical :: debug = .true.
                  
if (debug) write(*,'(80(1h+)/a)') 'subroutine newtsolver'

error = 0

! setup backstepping parameters

alf = 1.d-2
lambda = 1.d0 ! lambda determines how much of the newton corrections will be applied
lambdamin = 1.d-4
backstepping = .true.

! allocate working variables save current phi values for backstepping use

if (.not.allocated(phiold)) allocate(phiold(ptotal))
if (.not.allocated(delphi)) allocate(delphi(ptotal))

p = 0
do n = 1, allocatable_size(var_list(var_list_number(centring="all",type="independent"))%list)
  m = var_list(var_list_number(centring="all",type="independent"))%list(n)
  do ns = 1, ubound(var(m)%funk,1)
    p = p + 1
    phiold(p) = var(m)%funk(ns)%v
  end do
end do

! call the main linear solver routines

if (debug) write(*,*) 'calling mainsolver'
call time_process
call mainsolver(delphi,error)
call time_process(description='mainsolver')

! if there is a problem with the linear matrix solver then return

if (debug) write(*,*) 'in newtsolver after mainsolver, error = ',error
if (error /= 0) return

newtresold = newtres ! save newtres from last iteration for backstepping

backstepping_loop: do ! entrance point for repeat steps

! increment delvar changes to dependent variables

  p = 0
  do n = 1, allocatable_size(var_list(var_list_number(centring="all",type="independent"))%list)
    m = var_list(var_list_number(centring="all",type="independent"))%list(n)
    do ns = 1, ubound(var(m)%funk,1)
      p = p + 1
      var(m)%funk(ns)%v = phiold(p) + lambda*delphi(p)
    end do
  end do

! if backstepping is on, calculate new newtres value, and check that
!  it has decreased since newtresold

  call time_process
  call update
  call update_checker(error,newtres)
  call time_process(description='update')

! temp
! write(*,*) 'lambda, newtres = ',lambda, newtres
!stop

  if (backstepping) then

    if ( error /= 0 .or. & ! if update_checker detected an error
      (newtres > (1.d0-2.d0*alf*lambda)*newtresold.and.newtres > newtrestol) ) then
      lambda = lambda/2.d0
      if (debug.and.(newtres /= newtres)) write(*,'(a)') "BACKSTEPPING: newtres was NaN"
      if (debug.and.(error == 1)) write(*,'(a)') "BACKSTEPPING: update_check detected a NaN"
      if (debug.and.(error == 2)) write(*,'(a)') "BACKSTEPPING: update_check detected a dodgey variable property"
      if (lambda < lambdamin.and.error == 0) then
        if (debug) write(*,'(a,g8.3,a,g8.3)') "BACKSTEPPING: intermediate newtres = ",newtres, &
          ": lambda cannot be reduced further from lambda = ",lambda
        exit backstepping_loop
      end if
!     if (debug) write(*,'(a,g8.3,a,g8.3)') "BACKSTEPPING: intermediate newtres = ",newtres,": reducing lambda to ",lambda
      write(*,'(a,g8.3,a,g8.3)') "BACKSTEPPING: intermediate newtres = ",newtres,": reducing lambda to ",lambda
    else 
      if (debug) write(*,'(a,g8.3)') "BACKSTEPPING: exiting loop with lambda = ",lambda
      exit backstepping_loop ! residual has decreased from previous iteration (= newtres when lambda=0)
    end if

  else if ( error /= 0 ) then
    write(*,'(a)') "ERROR: update_checker has detected a NaN - try backstepping"
    stop
  else 
    exit backstepping_loop ! if not backstepping then exit backstepping loop
  end if

end do backstepping_loop

! deallocate(phiold,delphi) ! don't

if (debug) write(*,'(a/80(1h-))') 'subroutine newtsolver'

end subroutine newtsolver

!-----------------------------------------------------------------

subroutine mainsolver(delphi,error)

! here we setup sparse matrices to be used to produce one improvement in
!  the fields (delphi) using newton's method

use general_module
use equations_module
use intel_pardiso_module
use hsl_ma28d_module
use suitesparse_module

double precision, dimension(:), allocatable :: aa, delphi
integer, dimension(:), allocatable :: iaa, jaa
integer :: error, p, nz, nz_last_row, nz_col, nn
logical :: normalise = .false. ! true to normalise coefficients so that the largest row entry is positive 1
double precision :: aa_max
character(len=1000) :: formatline
logical :: manage_funk_dv_memory = .true. ! whether to deallocate and allocate dependent and equation dvs each time to minimum memory requirements
logical :: condition_number = .false. ! calculate the condition number of the matrix rather than solving the system
logical :: matrix_test = .false. ! do a test matrix inversion using a made-up test matrix instead of solving PDE system
logical :: debug = .false., debug_sparse = .true.

if (debug) debug_sparse = .true.
if (debug_sparse) write(*,'(80(1h+)/a)') 'subroutine mainsolver'

error = 1

if (manage_funk_dv_memory) then
  if (debug_sparse) write(*,*) 'deallocating dependent funk dvs'
  call memory_manage_dvs(type="dependent",action="deallocate")
end if

! sparse matrix produced is stored in compressed sparse row format (3 array variation), with 1 indexing (csr1)
! aa is value (length nz)
! jaa is column index (length nz)
! iaa is array of where new rows start (length n+1)

! calculate exact size that sparse matrix will be and size it
!if (debug_sparse) write(*,'(a)') 'sizing sparse matrix'
n = ptotal
nz = 0
do nn = 1, allocatable_size(var_list(var_list_number(centring="all",type="equation"))%list)
  m = var_list(var_list_number(centring="all",type="equation"))%list(nn)
  do ns = 1, ubound(var(m)%funk,1)
    nz = nz + var(m)%funk(ns)%ndv
  end do
end do
allocate(aa(nz),jaa(nz),iaa(n+1))
if (debug_sparse) write(*,'(a,i7,a,i10)') ' done sizing sparse matrix with n = ',n,': nz = ',nz

! calculate new aa and rhs (stored as delphi) matrix
! under newton's method, \tens{J} \dot \vect{\delta \phi} = - \vect{F}
! ie, aa = \tens{J}, rhs = -\vect{F}, rhs_returned = \vect{\delta \Phi}
! nz is the number of nonzero elements in aa
! n is the total number of equations (rank of matrix)

if (debug) write(80,*) 'looping through now: n = ',n

nz = 0
! loop through all equation lines
p = 0
do nn = 1, allocatable_size(var_list(var_list_number(centring="all",type="equation"))%list)
  m = var_list(var_list_number(centring="all",type="equation"))%list(nn)
  do ns = 1, ubound(var(m)%funk,1)
    p = p + 1

    delphi(p) = -var(m)%funk(ns)%v
    iaa(p) = nz + 1  ! row index
    if (debug) write(80,*) 'new row: p = ',p,': delphi(p) = ',delphi(p)
    if (debug) write(80,*) 'var(m)%funk(ns)%ndv = ',var(m)%funk(ns)%ndv,': var(m)%funk(ns)%pp = ',var(m)%funk(ns)%pp, &
      ': var(m)%funk(ns)%dv = ',var(m)%funk(ns)%dv

    nz_last_row = nz ! last entry in matrix of previous row is saved
    nz = nz + var(m)%funk(ns)%ndv
    if (nz>ubound(aa,1)) then
      write(*,*) 'ERROR: bounds of aa exceded in mainsolver: nz = ',nz
      stop
    end if
    jaa(nz_last_row+1:nz) = var(m)%funk(ns)%pp(1:var(m)%funk(ns)%ndv) ! column index (icn for hsl/coordinate, ja for intel storage)
    aa(nz_last_row+1:nz) = var(m)%funk(ns)%dv(1:var(m)%funk(ns)%ndv) ! matrix value

    if (normalise) then
      aa_max = 0.d0
      do nz_col = nz_last_row + 1, nz ! loop through aa entries on the current row (equation)
        if (abs(aa(nz_col)) > abs(aa_max)) aa_max = aa(nz_col) ! find one which has the largest magnitude
      end do
      if (debug) write(80,*) 'normalising row entries by aa_max = ',aa_max
      if (debug) write(80,*) 'final matrix entries:'
      do nz_col = nz_last_row + 1, nz
        if (abs(aa_max) > 1.d-20) aa(nz_col) = aa(nz_col)/aa_max
        if (debug) write(80,*) 'nz = ',nz_col,': aa(nz) = ',aa(nz_col)
      end do
      if (abs(aa_max) > 1.d-20) delphi(p) = delphi(p)/aa_max
      if (debug) write(80,*) 'delphi(p) = ',delphi(p)
    end if

  end do
end do

iaa(n+1) = nz + 1

! try test matrix as per intel manual rather than solving the real system
if (matrix_test) then
  nz = 13
  n = 5
  call resize_array(array=aa,new_size=13,keep_data=.false.)
  aa = [1, -1, -3, -2, 5, 4, 6, 4, -4, 2, 7, 8, -5] 
  call resize_array(array=jaa,new_size=13,keep_data=.false.)
  jaa = [1, 2, 4, 1, 2, 3, 4, 5, 1, 3, 4, 2, 5]
  call resize_array(array=iaa,new_size=6,keep_data=.false.)
  iaa = [1, 4, 6, 9, 12, 14]
  call resize_array(array=delphi,new_size=5,keep_data=.false.)
  delphi = 1.d0
  write(91,*) 'csr 1 basis matrix formed in mainsolver:'
  formatline = '(a,i3,a,g10.4,a,i3,a,i3)'
  do nn = 1, nz
    write(91,fmt=formatline) 'aa(',nn,') = ',aa(nn),': jaa(',nn,') = ',jaa(nn)
  end do
  formatline = '(a,i3,a,i3)'
  do nn = 1, n+1
    write(91,fmt=formatline) 'iaa(',nn,') = ',iaa(nn)
  end do
end if

if (manage_funk_dv_memory) then
  if (debug_sparse) write(*,*) 'deallocating equation funk dvs'
  call memory_manage_dvs(type="equation",action="deallocate")
end if

if (condition_number) then
  write(*,'(a)') 'SPECIAL: calling condition_number subroutine'
  call calc_condition_number(aa,iaa,jaa)
end if

if (debug_sparse) write(*,*) 'jacobian matrix has been formed, now calling linear solver'

! call appropriate third-party linear solver

if (trim(linear_solver) == "INTEL_PARDISO") then
! intel pardiso solver

  call intel_pardiso_linear_solver(aa,iaa,jaa,delphi,error)

else if (trim(linear_solver) == "INTEL_PARDISO_OOC") then
! intel pardiso solver out of core

  call intel_pardiso_linear_solver(aa,iaa,jaa,delphi,error,ooc=.true.)

else if (trim(linear_solver) == "SUITESPARSE_UMF") then
! suitesparse umf solver

  call suitesparse_umf_linear_solver(aa,iaa,jaa,delphi,error)

else if (trim(linear_solver) == "HSL_MA28") then
! hsl_ma28 solver

  call hsl_ma28_linear_solver(aa,iaa,jaa,delphi,error)

else
  write(*,*) 'ERROR: linear solver type '//trim(linear_solver)//' not known'
  stop
endif
if (error /= 0) write(*,*) 'WARNING: error from '//trim(linear_solver)//' linear solver: error = ',error

if (matrix_test) then
  write(91,*) 'solution returned in delphi'
  formatline = '(a,i3,a,g10.4)'
  do nn = 1, n
    write(91,fmt=formatline) 'delphi(',nn,') = ',delphi(nn)
  end do
  stop
end if

if (allocated(aa)) deallocate(aa)
if (allocated(iaa)) deallocate(iaa)
if (allocated(jaa)) deallocate(jaa)
  
if (manage_funk_dv_memory) then
  if (debug_sparse) write(*,*) 'reallocating dependent and equation funk dvs'
  call memory_manage_dvs(type="dependent",action="reallocate")
  call memory_manage_dvs(type="equation",action="reallocate")
end if

if (debug_sparse) write(*,'(a/80(1h-))') 'subroutine mainsolver'

end subroutine mainsolver

!-----------------------------------------------------------------

subroutine update_checker(error,newtres)

! here we run through all variables calculated by update checking that
!  all numbers are sensible (no NaNs)

use general_module
use equations_module

integer :: error, n, p, pp, mm, ij, ns, nvar, nnvar, m
double precision :: newtres
logical :: debug = .false.

if (debug) write(*,'(80(1h+)/a)') 'subroutine update_checker'

error = 1 ! assume there is a problem and instantly return if problem found

! check for NANs in equations and also find order of magnitude
p = 0
do nvar = 1, allocatable_size(var_list(var_list_number(centring="all",type="equation"))%list)
  m = var_list(var_list_number(centring="all",type="equation"))%list(nvar)
  do ns = 1, ubound(var(m)%funk,1)
    p = p + 1
    if (var(m)%funk(ns)%v /= var(m)%funk(ns)%v) return ! check value
    funkmag(p) = 0.d0
    do n = 1,var(m)%funk(ns)%ndv
      if (var(m)%funk(ns)%dv(n) /= var(m)%funk(ns)%dv(n)) return ! and dervatives
      pp = var(m)%funk(ns)%pp(n)
! search for independent variable that pp refers to
      do nnvar = 1, allocatable_size(var_list(var_list_number(centring="all",type="independent"))%list)
        mm = var_list(var_list_number(centring="all",type="independent"))%list(nnvar)
        if (var(mm)%funk(1)%pp(1) <= pp .and. var(mm)%funk(ubound(var(mm)%funk,1))%pp(1) >= pp) exit
      end do
      if (debug) then
        if (.not.(var(mm)%funk(1)%pp(1) <= pp .and. var(mm)%funk(ubound(var(mm)%funk,1))%pp(1) >= pp)) then
          write(*,*) 'ERROR: independent variable not found in update_checker'
          write(*,'(5(a,i6))') 'p = ',p,': m = ',m,': ns = ',ns,': pp = ',pp,': mm = ',mm
          stop
        end if
      end if
      funkmag(p) = max(funkmag(p),abs(var(m)%funk(ns)%dv(n)*var(mm)%magnitude))
      if (debug.and..false.) then
        write(*,'(5(a,i6))') 'p = ',p,': m = ',m,': ns = ',ns,': pp = ',pp,': mm = ',mm
        write(*,'(3(a,g14.6))') 'funkmag(p) = ',funkmag(p),': var(m)%funk(ns)%dv(n) = ', &
          var(m)%funk(ns)%dv(n),': var(mm)%magnitude = ',var(mm)%magnitude 
      end if
    end do
    if (debug) write(*,'(a,i6,a,g12.6,a)') &
      'funkmag(',p,') = ',funkmag(p),': var(m)%name = '//trim(var(m)%name)// &
      ': var(m)%funk(ns) = '//trim(print_funk(var(m)%funk(ns)))
    if (funkmag(p) < 1.d-20) then
      write(*,'(a,g14.6)') 'ERROR: the order of magnitude of '//trim(var(m)%type)//' '//trim(var(m)%name)// &
        ' is found to be = ',funkmag(p)
      write(*,'(a)') 'This indicates that this '//trim(var(m)%type)//' is not related to the dependent variables'
      write(*,'(a)') 'Equivalently, the partial derivatives of this '//trim(var(m)%type)//' with respect to the ', &
        'dependent variables are all zero'
      write(*,'(a)') 'Check that the variable equations which determine this '//trim(var(m)%type)//' are consistent and ', &
        'written in the correct top-down order'
      write(*,'(a)') 'Also check that none of these variables have the option ''noderivative'' set explicitly or ', &
        'implicitly (by being an output variable for example)'
      write(*,'(a)') 'If all else fails, debug this by setting debug = .true. in subroutine update_checker'
      stop
    end if
  end do
end do
  
! temp
!flush(6)

! check for NANs in dependents
do nvar = 1, allocatable_size(var_list(var_list_number(centring="all",type="dependent"))%list)
  m = var_list(var_list_number(centring="all",type="dependent"))%list(nvar)
  do ns = 1, ubound(var(m)%funk,1)
    if (var(m)%funk(ns)%v /= var(m)%funk(ns)%v) return ! check value
    do n = 1,var(m)%funk(ns)%ndv
      if (var(m)%funk(ns)%dv(n) /= var(m)%funk(ns)%dv(n)) return ! and dervatives
    end do
    if (debug) write(*,'(a)') 'var(m)%name = '//trim(var(m)%name)// &
      ': var(m)%funk(ns) = '//trim(print_funk(var(m)%funk(ns)))
  end do
end do

! check for NANs in independents
do nvar = 1, allocatable_size(var_list(var_list_number(centring="all",type="independent"))%list)
  m = var_list(var_list_number(centring="all",type="independent"))%list(nvar)
  do ns = 1, ubound(var(m)%funk,1)
    if (var(m)%funk(ns)%v /= var(m)%funk(ns)%v) return ! check value
    do n = 1,var(m)%funk(ns)%ndv
      if (var(m)%funk(ns)%dv(n) /= var(m)%funk(ns)%dv(n)) return ! and dervatives
    end do
    if (debug) write(*,'(a)') 'var(m)%name = '//trim(var(m)%name)// &
      ': var(m)%funk(ns) = '//trim(print_funk(var(m)%funk(ns)))
  end do
end do

error = 0

! check that certain properties of the variables are maintained
call variable_checker(error) 

call residual(newtres)

if (debug) write(*,'(a/80(1h-))') 'subroutine update_checker'

end subroutine update_checker

!-----------------------------------------------------------------

subroutine residual(res)

! finds residual

use general_module
use equations_module
double precision :: res, aatmp, aatmpmax
integer :: ns, m, p, normtype, nserr, merr, perr, nvar
logical :: debug = .false.

if (debug) write(*,'(80(1h+)/a)') 'subroutine residual'

normtype = 2 ! norm type to use (0=infinity norm, 1=abs average norm, 2=rms norm)

res = 0.d0

aatmpmax = 0.d0
nserr = -1
merr = -1
perr = -1

p = 0
do nvar = 1, allocatable_size(var_list(var_list_number(centring="all",type="equation"))%list)
  m = var_list(var_list_number(centring="all",type="equation"))%list(nvar)
  do ns = 1, ubound(var(m)%funk,1)
    p = p + 1
    aatmp = abs(var(m)%funk(ns)%v)/funkmag(p)
    if (normtype == 0) then
      res = max(res,aatmp)  ! infinity norm
    else if (normtype.eq.1) then
      res = res + aatmp ! 1 norm
    else
      res = res + aatmp**2 ! 2 norm
    endif
! also record maximum error location
    if (aatmp.gt.aatmpmax) then
      aatmpmax = aatmp
      merr = m
      nserr = ns
      perr = p
    endif
  end do
end do

if (normtype == 1.or.normtype == 2) res = res/dble(ptotal)

if (normtype == 2) res = sqrt(res)

if (debug) write(*,*) 'maximum error is occuring at ns = ',nserr,': m = ',merr,'('//trim(var(merr)%name)// &
  ') : p = ',perr,': funk = ',var(merr)%funk(nserr)%v,': and funkmag = ',funkmag(perr)

if (debug) write(*,'(a/80(1h-))') 'subroutine residual'

end subroutine residual

!-----------------------------------------------------------------

subroutine calc_condition_number(aa,iaa,jaa)

! diagnostic routine to calculate the condition number of the jacobian matrix using
!  svd
! requires lots of memory!

use lapack_module

double precision, dimension(:) :: aa ! already allocated
integer, dimension(:) :: iaa, jaa ! already allocated
double precision, dimension(:,:), allocatable :: aa_full
double precision, dimension(:), allocatable :: w
double precision, dimension(:,:), allocatable :: u, v
double precision :: wmin, wmax
integer :: n, nz, nn, nr
logical :: error

! arrays produced by the mainsolver subroutine are in compressed sparse row (csr) format with 1 indexing
! find array bounds and check that they are consistent
n = ubound(iaa,1) - 1
nz = iaa(n+1) - 1
write(60,*) 'n = ',n,': nz = ',nz
if (ubound(aa,1) /= nz.or.ubound(jaa,1) /= nz) stop &
  'ERROR: some of the csr matrices that are entering calc_condition_number are incorrectly sized'

! first thing is allocate all variables and reconstruct aa matrix using nonsparse storage from csr1
allocate(aa_full(n,n),u(n,n),w(n),v(n,n))
aa_full = 0.d0
u = 0.d0
w = 0.d0
v = 0.d0
do nr = 1, n
  if (iaa(nr) < iaa(nr+1)) then
    do nn = iaa(nr), iaa(nr+1)-1
      aa_full(nr,jaa(nn)) = aa(nn)
    end do
  end if
end do
  
! now do svd of aa_full
call lapack_singular_value_decomposition(a=aa_full,u=u,w=w,v=v,error=error)

if (error) then
  write(*,*) 'ERROR: problem doing svd in condition_number'
  stop
end if
  
do nn = 1, n
  write(60,*) 'nn = ',nn,': w(nn) = ',w(nn)
end do
wmin = minval(w)
wmax = maxval(w)
write(60,*) 'wmin = ',wmin,': wmax = ',wmax,': wmax/wmin = ',wmax/wmin
write(*,'(a,g12.5)') 'SPECIAL: condition number = ',wmax/wmin

end subroutine calc_condition_number

!-----------------------------------------------------------------

end module solver_module
!-----------------------------------------------------------------
